# Based on: https://github.com/python/cpython/blob/master/Mac/BuildScript/resources/install_certificates.command

# install_certifi.py
#
# sample script to install or update a set of default Root Certificates
# for the ssl module.  Uses the certificates provided by the certifi package:
#       https://pypi.org/project/certifi/

import contextlib
import os
import os.path
import ssl
import stat
import subprocess
import sys

STAT_0o775 = (
    stat.S_IRUSR
    | stat.S_IWUSR
    | stat.S_IXUSR
    | stat.S_IRGRP
    | stat.S_IWGRP
    | stat.S_IXGRP
    | stat.S_IROTH
    | stat.S_IXOTH
)


def main() -> None:
    openssl_dir, openssl_cafile = os.path.split(ssl.get_default_verify_paths().openssl_cafile)
    print(" -- pip install --upgrade certifi")
    subprocess.check_call(
        [sys.executable, "-E", "-s", "-m", "pip", "install", "--upgrade", "certifi"]
    )

    import certifi  # noqa: PLC0415

    # change working directory to the default SSL directory
    os.chdir(openssl_dir)
    relpath_to_certifi_cafile = os.path.relpath(certifi.where())

    print(" -- removing any existing file or link")
    with contextlib.suppress(FileNotFoundError):
        os.remove(openssl_cafile)
    print(" -- creating symlink to certifi certificate bundle")
    os.symlink(relpath_to_certifi_cafile, openssl_cafile)

    print(" -- setting permissions")
    os.chmod(openssl_cafile, STAT_0o775)
    print(" -- update complete")


if __name__ == "__main__":
    main()
