
/* Global styles */

body {
    overflow-wrap: break-word;
    overflow-x: hidden;
    -webkit-font-smoothing: antialiased;
}

p {
    margin-bottom: 12px;
    line-height: 26px;
}

pre {
  margin: 1em 0;
}

pre:first-child {
  margin-top: 0;
}

code {
    font-size: 85%;
    background-color: rgba(27,31,35,.05);
    border-radius: 3px;
    padding: .15em .4em;
    color: inherit;
    border: none;
    margin: 0 1px;
}

code, pre, tt {
    font-family: SFMono-Regular, Consolas, Liberation Mono, Menlo, monospace;
    line-height: 1.4;
}

pre code, pre code.hljs {
  padding: 1em;
  background-color: #f6f8fa;
}

h1, h2, h3, h4, h5, h6 {
    margin-top: 2em;
    border-bottom: 1px solid  #e1e4e5;
    margin-bottom: 0.8em;
    padding-bottom: 0.2em;
    scroll-margin: 1.5em 0;
}

/* make it so tables can overflow the content area, make the most of the content area */
.wy-nav-content-wrap {
  container-type: inline-size;
}
.wy-table-responsive {
  overflow-x: auto;
  overflow-y: hidden;
  position: relative;
  left: -3.236em;
  max-width: calc(100cqw);
  width: fit-content;
  min-width: min(calc(100% + 3.236em + 3.236em), 100cqw);
}
.rst-content .section .wy-table-responsive .docutils {
  padding-left: 3.236em;
  padding-right: 3.236em;
}
.rst-content .section .wy-table-responsive .docutils thead,
.rst-content .section .wy-table-responsive .docutils tbody {
  background-color: white;
}
@media screen and (max-width: 768px) {
  .wy-table-responsive {
    left: -1em;
  }
  .rst-content .section .wy-table-responsive .docutils {
    padding-left: 1em;
    padding-right: 1em;
  }
}

.wy-nav-content {
  max-width: 900px;
}

.rst-content blockquote {
    margin-bottom: 14px;
}

.rst-content th {
    white-space: normal;
}

.rst-content .admonition {
  line-height: inherit;
}

.admonition code {
  border: none;
}

@media screen and (max-width: 768px) {
  .wy-nav-content-wrap .wy-nav-content {
      padding: 1em;
  }
}


/*
    Code block filename style

    Use by putting a quote before a code block e.g.

       > filename.txt

       ```
       code here
       ```

    JS adds the relevant classes to make the following selectors work.
*/
.rst-content .code-block-filename {
    margin: 0;
    padding: 2px 1em;
    background-color: #eff1f3;
    font-size: 85%;
    font-weight: bold;

    border-top-left-radius: 3px;
    border-top-right-radius: 3px;

    border-bottom: 1px solid #e6e8e9;
    color: #444;
}

.rst-content .code-block-filename * {
    font-size: inherit;
}
.rst-content .code-block-filename :last-child {
    margin-bottom: 0;
}

.code-block-filename + pre {
    margin-top: 0;
    border-top-left-radius: 0;
    border-top-right-radius: 0;
}

.code-block-filename + pre code {
    padding-top: 0.9em;
}
/*
.toctree-l3 {
  border-left: 10px solid transparent;
} */

/* import font awesome 4 for icons */
@import url(https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css);


/* Tabs */

/* Style the tab */
.tabs {
  margin-bottom: 1em;
  -webkit-font-smoothing: antialiased;
}

.tabs-header {
  background-color: white;
  display: flex;
  justify-content: flex-start;
  overflow-x: auto;
  overflow-y: hidden;
  position: relative;
}
.tabs-header::before {
  position: absolute;
  content: "";
  bottom: 0;
  left: 0;
  right: 0;
  height: 2px;
  background: #ccc;
}
.tabs-header > * {
  position: relative;
}

/* Style the buttons that are used to open the tab content */
.tabs-header button {
  background-color: transparent;
  border: none;
  outline: none;
  cursor: pointer;
  color: inherit;
  font: inherit;
  font-size: 0.9em;
  font-weight: 600;
  padding: 8px 16px;
  border-bottom: 2px solid #ccc;
}

.tabs-header button:focus-visible {
  /* preserve an outline for accessibility purposes */
  outline: 1px solid currentColor;
  outline-offset: -1px;
}

/* Change background color of buttons on hover */
.tabs-header button:hover {
  background-color: rgba(0, 0, 0, 0.02);
}

/* Create an active/current tablink class */
.tabs-header button.active {
  color: #4086bd;
  border-bottom-color: currentColor;
}

/* Style the tab content */
.tabs-content {
  background-color: #f1f6fa;
  padding: 1px 1em;
  padding-top: 0.8em;

  /* don't collapse inner margins */
  overflow-y: hidden;
}

/* Examples tabs styling */

.tabs.examples .tabs-header {
  justify-content: flex-end;
  background-color: transparent;
  border-bottom-style: none;
  padding: 0;
}
.tabs.examples .tabs-header::before {
  height: 0;
}
.tabs.examples .tabs-header button {
  padding: 6px 12px;
  border-bottom: none;
  color: rgba(0, 0, 0, 0.3);
  font-size: 0.8em;
  border-top-left-radius: 4px;
  border-top-right-radius: 4px;
}
.tabs-header button:hover {
  background-color: unset;
}

.tabs.examples .tabs-header button.active {
  border-bottom-style: none;
  background: #f6f8fa;
  color: inherit;
}
.tabs.examples .tabs-content {
  padding: 0;
  background-color: transparent;
  background-color: #f6f8fa;
}
.tabs.examples .tabs-content .tab > * {
  padding-left: 0.8em;
  padding-right: 0.8em;
}
.tabs.examples .tabs-content .tab > pre {
  padding-left: 0;
  padding-right: 0;
  margin-bottom: 0.5em;
}

/* mkdocs update theme fix-ups */

.rst-content code, .rst-content tt, code {
  /* border: none; */
  /* background-color: #f0f1f1; */
  color: inherit;
}
.rst-content pre code {
  border: none;
}

.rst-content h1 code,
.rst-content h2 code,
.rst-content h3 code,
.rst-content h4 code,
.rst-content h5 code,
.rst-content h6 code {
  border: none;
  color: inherit;
  background-color: #f0f1f1;
  font-size: 80%;
}

.rst-content table.docutils td code,
.rst-content table.docutils th code,
.rst-content table.field-list td code,
.rst-content table.field-list th code,
.wy-table td code,
.wy-table th code {
  /* table elements are already made smaller, the code styling on top of that makes the text too small */
  font-size: 82.5%;
}


/* hide the l1 buttons */
.wy-menu-vertical li.toctree-l1.current>a button.toctree-expand,
.wy-menu-vertical li.toctree-l1.on>a button.toctree-expand,
.wy-menu-vertical li.toctree-l1>a button.toctree-expand {
  display: none;
}


/* word wrap in table cells */
.wy-table-responsive table td, .wy-table-responsive table th {
  white-space: normal;
  line-height: 1.4;
}
